/* Sieve of Eratosthenes, based on an old C* example.          */
/* See C* Reference Manual, Version 4.0A Field Test, TMC, 1987 */
/* File:  sieve.cs					       */

#include <stdio.h>
#include <stdlib.h>
#include <cscomm.h> 
#include <cm/cmfs.h>
#include <cm/cm_file.h>

/* Size for sieve calculations, 32K is convenient for demos */
#define CSIZE 32768                        /* column size */ 

/* Boolean constants */
#define TRUE  1
#define FALSE 0

/* Other constants */
#define RSIZE 16384                        /* row size */ 
 
shape [CSIZE] sieve;	/* Shape for sieve representation */

int:sieve  number;	/* Number represented by this proc. */
bool:sieve is_prime;	/* True if number is prime */
bool:sieve mk[RSIZE];   /* bit vector with primes marked as true */
int:sieve i;            /* Represents pcoord */ 
int:sieve row_i;        /* Number representing #row x pcoord */
bool:sieve x_prime;

int limit;	/* Limit on sieve calculation, < CSIZE */

/* Find all primes <= limit */
/* Assumes  [0]number = x for x = 3,5,7,9, .... */

void find_primes()
 { int next_prime;		/* Next prime identified */
   bool:sieve is_candidate;     /* True of proc's that might be prime */
   with (sieve)
    { /* Initialize */
      is_prime = FALSE;
      is_candidate = (number <= limit);
      next_prime = 1;  /* To start the loop */
      /* Loop continues until all candidates are prime */
      while ( next_prime*next_prime < limit )
       where (is_candidate)
	 { next_prime = (<?= number);
	   where ( (number % next_prime) == 0 ) is_candidate = FALSE;
           where (number == next_prime) is_prime = TRUE;	
        /* Mark next_prime as prime */
	 }
      /* All remaining candidates are prime */
      where (is_candidate) is_prime = TRUE;
      where (number == 2) is_prime = FALSE;
    }  /* end with */
 }

/* Print (some of) the primes, 10 per line.		*/
/* pall:  -1 - none,  0 - in last 100 numbers,  1- all	*/

void print_primes(int pall)
 { int j, nx;
   bool:sieve pprime;	/* Identifies primes to be printed */
   j = 1;
   with (sieve)
    { if ( pall >= 0 ) pprime = is_prime; else pprime = FALSE;
      if ( pall == 0 ) pprime &= (number >= (limit - 100));
      /* Loop continues until desired primes are printed */
      /* This amounts to "do-for-selected-processors"	 */
      while (|= pprime)
       where (pprime)
	{ nx = (<?= number);
          where (number == nx) pprime = FALSE;
	  /* Note that nx processed */
	  printf("%6d ", nx);
	  if ( (j++) == 10 ) { printf("\n");  j = 1; };
	} /* end where, end while */
      if (j != 1) printf("\n");
    } /* end with */
 }

/* Prints out at a starting number(num), 25 number and see if it is */   
/* markoff or out */

check_mk(int num)
{
  int col_p,row_p;
  int n,c,tn,j; 

  /* initialization */
  c = 0;
  n = num % 100;
  col_p = (num-3)/(RSIZE * 2);
  row_p = ((num-3)/2) % RSIZE;
  tn = row_p + 20;

  printf("Bits for odd integers starting at %10d \n",num);

  /* printing the numbers that are being checked */  
  for (j=0; j<20; j++)
   { printf("%3d",n);
     if (n == 99) n = -1; 
     n = n + 2; } 
  printf("\n");

   /* looping to print the mark in mk for checking */ 
  for (; row_p < tn; row_p++,c++)
   { 
     if (row_p > (RSIZE-1))
      {  col_p++;
         row_p = 0;
         tn = row_p+(20-c);
         c = 0;
       } /* end if */ 
     printf("%3d",[col_p]mk[row_p]); 
   } /* end for */ 
  printf("\n"); 
}  

void copy_ispr()
 { int:sieve c,ck;
   int r; 

   with(sieve)
    { c = pcoord(0);
      x_prime = 0;
      ck = 1 + c * RSIZE;

      where (c < 2) {
       for (r = 0; r < RSIZE; r++)
        { [ck]x_prime = mk[r];
          ck++; }
       }  /* end where */
      [0]x_prime = 0; 
    } /* end with */
} 

 
void compare()
{ int ct;
  int k = 0;   
  ct = 0;

  with (sieve)
   {  while (ct < CSIZE)
       { if ( [ct]x_prime != [ct]is_prime)
          { printf("[%5d]x_prime = %2d          ",ct,[ct]x_prime);
            printf("[%5d]is_prime = %2d\n",ct,[ct]is_prime);
          }     
         ct++;  
       }  /* end while */
    } /* end with */
printf("\n");  
} 

void num_errors()
 {  int:sieve cm;
    int:sieve num;
    int totnum;
    
    totnum = 0;  
    num = 0;

    with (sieve)
     {  cm = pcoord(0);
        where ([cm]x_prime != [cm]is_prime) num++;
      }
    totnum += num; 
    printf("The number of errors is %5d\n",totnum);
}

void check_div_mk(int many, int num, int cl, int rw) 
 {  bool nprime, error;
    int div,counter;
    bool:sieve flag;

    error = 0;
    counter = 0;

    printf("many = %6d, num = %11d, cl = %6d, rw = %6d\n",many,num,cl,rw);
    find_primes();

    with (sieve)
    { while (!error && (counter < many))   
      { 
         flag = (((num % number) == 0) & (num != number));
         nprime = !(|= flag);
         if (nprime == 0) 
          { where (flag)  div = (<?= number); } 

         if (nprime == 0)
           {  if ([cl]mk[rw] != nprime)
               { error++;
                 printf("\n"); 
                 printf(" ***** ERROR ***** \n");
                 printf("number = %11d divisible by %6d \n",num,div);
                 printf("[%6d]mk[%6d] = %2d\n",cl,rw,[cl]mk[rw]);
                 printf("should be 0\n");
               }   /* end if */
            } /* end if */
         else if (nprime == 1)
           { if ([cl]mk[rw] != nprime)
               { error++;
                 printf("\n"); 
                 printf(" ***** ERROR ***** \n");
                 printf("number = %11d \n",num);
                 printf("[%6d]mk[%6d] = %2d\n",cl,rw,[cl]mk[rw]);
                 printf("should be 1\n");
               }   /* end if */
            } /* end if */

         num = num + 2;
         rw++;  counter++;
         if (rw == RSIZE) { rw = 0; cl++; }
       } /* end while */
    } /* end with */
    printf("\n");  
    printf("num = %11d, cl = %6d, rw = %6d\n",num,cl,rw);  
}  
  
void checkpoint()
 {
   int mark,ispr,dvfd;
   char fnm[50], type[2]; 

   ispr = boolsizeof(bool:sieve);
   mark = RSIZE * ispr;

   printf("r,w> ");  scanf("%s",type);
   printf("Checkpoint file> ");  scanf("%s", fnm);  

   with (sieve)  
   { if ( ( type[0] == 'w' ) && ( type[1] == 0 ) )
      { /* Checkpoint in user-specified files */
        CM_start_timer(1);
        dvfd = CMFS_creat(fnm, 0600);
        CMFS_write_file(dvfd, &mk[0], mark); 
        CMFS_write_file(dvfd, &is_prime, ispr);  
        CMFS_close(dvfd);
        CM_stop_timer(1);  CM_reset_timer();
      }
    else if ( ( type[0] == 'r' ) && ( type[1] == 0 ) )
      { /* Resume from user-specified files */
        CM_start_timer(1);
        dvfd = CMFS_open(fnm, CMFS_O_RDONLY);
        CMFS_read_file(dvfd, &mk[0], mark);  
        CMFS_read_file(dvfd, &is_prime,  ispr);  
        CMFS_close(dvfd);
        CM_stop_timer(1);  CM_reset_timer();
      }
   }
}

/* Main program.  Commands are:
   d <int>  - display status: -1 just time, 0 in last 100, 1 all
   l <int>  - set limit to int
   c        - run compare routine  
   k <int>  - starting checking at <int> an odd integer   
   o        - run copy_ispr routine  
   e        - run num_errors routine 
   h <int,int,int,int>  - check_div_mk() giving
                    <how many num check, num start with,column, row>
   x        - checkpointing
   p        - find primes
   q        _ quit           */

void main()
{  int display;  char c[2];
   int n,j,mn,ns,cm,ro; 
   display = -1; limit = 65535;  

    /* Initialize `number' for integers 2, 3, .... */
    with (sieve) number = pcoord(0)*2+1;
    [0]number = 2;

    /* initialize mk,i and row_i */
    with (sieve) 
     { for (j=0; j<RSIZE; j++) mk[j] = TRUE;
        i = pcoord(0);
        row_i = RSIZE * i;  
     } /* end of with */

    while ( printf("sieve>"), scanf("%1s", c), c[0] != 'q' )
     switch (c[0])
      { case 'd': scanf("%d", &display); break;
        case 'l': scanf("%d", &limit);  break;
        case 'p':
	  printf("Calculating primes ...\n");
	  CM_start_timer(1);
		find_primes();
	  CM_stop_timer(1);  CM_reset_timer();
                print_primes(display);  
	  break;
        case 'h': scanf("%d %d %d %d",&mn,&ns,&cm,&ro); 
          CM_start_timer(1);
                check_div_mk(mn,ns,cm,ro);  
          CM_stop_timer(1); CM_reset_timer();
          break; 
        case 'c': 
          CM_start_timer(1);
                compare();
          CM_stop_timer(1); CM_reset_timer();
          break;
        case 'o': 
          CM_start_timer(1);
                copy_ispr();
          CM_stop_timer(1); CM_reset_timer();
          break;
        case 'x': 
          checkpoint(); 
          break; 
        case 'k': scanf("%d",&n);
          check_mk(n);
          break;
        case 'e':
          CM_start_timer(1);
                num_errors();
          CM_stop_timer(1); CM_reset_timer();
          break;
	default:   printf("Bad input - ignored\n"); break;
      }
 }







































































